@@ -22,33 +22,51 @@
 
 #include "config.h"
 
+#include <stdio.h>
 #include <stdlib.h>
 #include <stdint.h>
 #include <limits.h>
 #include <math.h>
 
 #include "filter.h"
+#include "libinput-util.h"
 
 void
 filter_dispatch(struct motion_filter *filter,
 		struct motion_params *motion,
-		void *data, uint32_t time)
+		void *data, uint64_t time)
 {
 	filter->interface->filter(filter, motion, data, time);
 }
 
+void
+filter_destroy(struct motion_filter *filter)
+{
+	if (!filter)
+		return;
+
+	filter->interface->destroy(filter);
+}
+
 /*
- * Pointer acceleration filter
+ * Default parameters for pointer acceleration profiles.
  */
 
-#define MAX_VELOCITY_DIFF	1.0
+#define DEFAULT_THRESHOLD 0.4			/* in units/ms */
+#define DEFAULT_ACCELERATION 2.0		/* unitless factor */
+
+/*
+ * Pointer acceleration filter constants
+ */
+
+#define MAX_VELOCITY_DIFF	1.0 /* units/ms */
 #define MOTION_TIMEOUT		300 /* (ms) */
 #define NUM_POINTER_TRACKERS	16
 
 struct pointer_tracker {
-	double dx;
-	double dy;
-	uint32_t time;
+	double dx;	/* delta to most recent event, in device units */
+	double dy;	/* delta to most recent event, in device units */
+	uint64_t time;  /* ms */
 	int dir;
 };
 
@@ -58,77 +76,19 @@ struct pointer_accelerator {
 
 	accel_profile_func_t profile;
 
-	double velocity;
-	double last_velocity;
-	int last_dx;
-	int last_dy;
+	double velocity;	/* units/ms */
+	double last_velocity;	/* units/ms */
+	int last_dx;		/* device units */
+	int last_dy;		/* device units */
 
 	struct pointer_tracker *trackers;
 	int cur_tracker;
 };
 
-enum directions {
-	N  = 1 << 0,
-	NE = 1 << 1,
-	E  = 1 << 2,
-	SE = 1 << 3,
-	S  = 1 << 4,
-	SW = 1 << 5,
-	W  = 1 << 6,
-	NW = 1 << 7,
-	UNDEFINED_DIRECTION = 0xff
-};
-
-static int
-get_direction(int dx, int dy)
-{
-	int dir = UNDEFINED_DIRECTION;
-	int d1, d2;
-	double r;
-
-	if (abs(dx) < 2 && abs(dy) < 2) {
-		if (dx > 0 && dy > 0)
-			dir = S | SE | E;
-		else if (dx > 0 && dy < 0)
-			dir = N | NE | E;
-		else if (dx < 0 && dy > 0)
-			dir = S | SW | W;
-		else if (dx < 0 && dy < 0)
-			dir = N | NW | W;
-		else if (dx > 0)
-			dir = NW | W | SW;
-		else if (dx < 0)
-			dir = NE | E | SE;
-		else if (dy > 0)
-			dir = SE | S | SW;
-		else if (dy < 0)
-			dir = NE | N | NW;
-	}
-	else {
-		/* Calculate r within the interval  [0 to 8)
-		 *
-		 * r = [0 .. 2π] where 0 is North
-		 * d_f = r / 2π  ([0 .. 1))
-		 * d_8 = 8 * d_f
-		 */
-		r = atan2(dy, dx);
-		r = fmod(r + 2.5*M_PI, 2*M_PI);
-		r *= 4*M_1_PI;
-
-		/* Mark one or two close enough octants */
-		d1 = (int)(r + 0.9) % 8;
-		d2 = (int)(r + 0.1) % 8;
-
-		dir = (1 << d1) | (1 << d2);
-	}
-
-	return dir;
-}
-
 static void
 feed_trackers(struct pointer_accelerator *accel,
 	      double dx, double dy,
-	      uint32_t time)
+	      uint64_t time)
 {
 	int i, current;
 	struct pointer_tracker *trackers = accel->trackers;
@@ -144,7 +104,7 @@ feed_trackers(struct pointer_accelerator *accel,
 	trackers[current].dx = 0.0;
 	trackers[current].dy = 0.0;
 	trackers[current].time = time;
-	trackers[current].dir = get_direction(dx, dy);
+	trackers[current].dir = vector_get_direction(dx, dy);
 }
 
 static struct pointer_tracker *
@@ -157,7 +117,7 @@ tracker_by_offset(struct pointer_accelerator *accel, unsigned int offset)
 }
 
 static double
-calculate_tracker_velocity(struct pointer_tracker *tracker, uint32_t time)
+calculate_tracker_velocity(struct pointer_tracker *tracker, uint64_t time)
 {
 	int dx;
 	int dy;
@@ -166,37 +126,24 @@ calculate_tracker_velocity(struct pointer_tracker *tracker, uint32_t time)
 	dx = tracker->dx;
 	dy = tracker->dy;
 	distance = sqrt(dx*dx + dy*dy);
-	return distance / (double)(time - tracker->time);
+	return distance / (double)(time - tracker->time); /* units/ms */
 }
 
 static double
-calculate_velocity(struct pointer_accelerator *accel, uint32_t time)
+calculate_velocity(struct pointer_accelerator *accel, uint64_t time)
 {
 	struct pointer_tracker *tracker;
 	double velocity;
 	double result = 0.0;
-	double initial_velocity;
+	double initial_velocity = 0.0;
 	double velocity_diff;
 	unsigned int offset;
 
 	unsigned int dir = tracker_by_offset(accel, 0)->dir;
 
-	/* Find first velocity */
-	for (offset = 1; offset < NUM_POINTER_TRACKERS; offset++) {
-		tracker = tracker_by_offset(accel, offset);
-
-		if (time <= tracker->time)
-			continue;
-
-		result = initial_velocity =
-			calculate_tracker_velocity(tracker, time);
-		if (initial_velocity > 0.0)
-			break;
-	}
-
 	/* Find least recent vector within a timelimit, maximum velocity diff
 	 * and direction threshold. */
-	for (; offset < NUM_POINTER_TRACKERS; offset++) {
+	for (offset = 1; offset < NUM_POINTER_TRACKERS; offset++) {
 		tracker = tracker_by_offset(accel, offset);
 
 		/* Stop if too far away in time */
@@ -211,27 +158,31 @@ calculate_velocity(struct pointer_accelerator *accel, uint32_t time)
 
 		velocity = calculate_tracker_velocity(tracker, time);
 
-		/* Stop if velocity differs too much from initial */
-		velocity_diff = fabs(initial_velocity - velocity);
-		if (velocity_diff > MAX_VELOCITY_DIFF)
-			break;
+		if (initial_velocity == 0.0) {
+			result = initial_velocity = velocity;
+		} else {
+			/* Stop if velocity differs too much from initial */
+			velocity_diff = fabs(initial_velocity - velocity);
+			if (velocity_diff > MAX_VELOCITY_DIFF)
+				break;
 
-		result = velocity;
+			result = velocity;
+		}
 	}
 
-	return result;
+	return result; /* units/ms */
 }
 
 static double
 acceleration_profile(struct pointer_accelerator *accel,
-		     void *data, double velocity, uint32_t time)
+		     void *data, double velocity, uint64_t time)
 {
 	return accel->profile(&accel->base, data, velocity, time);
 }
 
 static double
 calculate_acceleration(struct pointer_accelerator *accel,
-		       void *data, double velocity, uint32_t time)
+		       void *data, double velocity, uint64_t time)
 {
 	double factor;
 
@@ -246,39 +197,18 @@ calculate_acceleration(struct pointer_accelerator *accel,
 
 	factor = factor / 6.0;
 
-	return factor;
-}
-
-static double
-soften_delta(double last_delta, double delta)
-{
-	if (delta < -1.0 || delta > 1.0) {
-		if (delta > last_delta)
-			return delta - 0.5;
-		else if (delta < last_delta)
-			return delta + 0.5;
-	}
-
-	return delta;
-}
-
-static void
-apply_softening(struct pointer_accelerator *accel,
-		struct motion_params *motion)
-{
-	motion->dx = soften_delta(accel->last_dx, motion->dx);
-	motion->dy = soften_delta(accel->last_dy, motion->dy);
+	return factor; /* unitless factor */
 }
 
 static void
 accelerator_filter(struct motion_filter *filter,
 		   struct motion_params *motion,
-		   void *data, uint32_t time)
+		   void *data, uint64_t time)
 {
 	struct pointer_accelerator *accel =
 		(struct pointer_accelerator *) filter;
-	double velocity;
-	double accel_value;
+	double velocity; /* units/ms */
+	double accel_value; /* unitless factor */
 
 	feed_trackers(accel, motion->dx, motion->dy, time);
 	velocity = calculate_velocity(accel, time);
@@ -287,8 +217,6 @@ accelerator_filter(struct motion_filter *filter,
 	motion->dx = accel_value * motion->dx;
 	motion->dy = accel_value * motion->dy;
 
-	apply_softening(accel, motion);
-
 	accel->last_dx = motion->dx;
 	accel->last_dy = motion->dy;
 
@@ -332,3 +260,51 @@ create_pointer_accelator_filter(accel_profile_func_t profile)
 
 	return &filter->base;
 }
+
+static inline double
+calc_penumbral_gradient(double x)
+{
+	x *= 2.0;
+	x -= 1.0;
+	return 0.5 + (x * sqrt(1.0 - x * x) + asin(x)) / M_PI;
+}
+
+double
+pointer_accel_profile_smooth_simple(struct motion_filter *filter,
+				    void *data,
+				    double velocity, /* units/ms */
+				    uint64_t time)
+{
+	double threshold = DEFAULT_THRESHOLD; /* units/ms */
+	double accel = DEFAULT_ACCELERATION; /* unitless factor */
+	double smooth_accel_coefficient; /* unitless factor */
+	double factor; /* unitless factor */
+
+	if (threshold < 0.1)
+		threshold = 0.1;
+	if (accel < 1.0)
+		accel = 1.0;
+
+	/* We use units/ms as velocity but it has no real meaning unless all
+	   devices have the same resolution. For touchpads, we normalize to
+	   400dpi (15.75 units/mm), but the resolution on USB mice is all
+	   over the place. Though most mice these days have either 400
+	   dpi (15.75 units/mm), 800 dpi or 1000dpi, excluding gaming mice
+	   that can usually adjust it on the fly anyway and currently go up
+	   to 8200dpi.
+	  */
+	if (velocity < (threshold / 2.0))
+		return calc_penumbral_gradient(0.5 + velocity / threshold) * 2.0 - 1.0;
+
+	if (velocity <= threshold)
+		return 1.0;
+
+	factor = velocity/threshold;
+	if (factor >= accel)
+		return accel;
+
+	/* factor is between 1.0 and accel, scale this to 0.0 - 1.0 */
+	factor = (factor - 1.0) / (accel - 1.0);
+	smooth_accel_coefficient = calc_penumbral_gradient(factor);
+	return 1.0 + (smooth_accel_coefficient * (accel - 1.0));
+}
