@@ -25,8 +25,10 @@
 
 #include "config.h"
 
+#include <stdint.h>
+
 struct motion_params {
-	double dx, dy;
+	double dx, dy; /* in units/ms @ 400dpi */
 };
 
 struct motion_filter;
@@ -34,13 +36,15 @@ struct motion_filter;
 void
 filter_dispatch(struct motion_filter *filter,
 		struct motion_params *motion,
-		void *data, uint32_t time);
+		void *data, uint64_t time);
+void
+filter_destroy(struct motion_filter *filter);
 
 
 struct motion_filter_interface {
 	void (*filter)(struct motion_filter *filter,
 		       struct motion_params *motion,
-		       void *data, uint32_t time);
+		       void *data, uint64_t time);
 	void (*destroy)(struct motion_filter *filter);
 };
 
@@ -54,9 +58,24 @@ create_linear_acceleration_filter(double speed);
 typedef double (*accel_profile_func_t)(struct motion_filter *filter,
 				       void *data,
 				       double velocity,
-				       uint32_t time);
+				       uint64_t time);
 
 struct motion_filter *
 create_pointer_accelator_filter(accel_profile_func_t filter);
 
+
+/*
+ * Pointer acceleration profiles.
+ */
+
+/*
+ * Profile similar which is similar to nonaccelerated but with a smooth
+ * transition between accelerated and non-accelerated.
+ */
+double
+pointer_accel_profile_smooth_simple(struct motion_filter *filter,
+				    void *data,
+				    double velocity,
+				    uint64_t time);
+
 #endif /* FILTER_H */
