@@ -33,6 +33,7 @@
 #include "libinput.h"
 #include "libinput-private.h"
 #include "evdev.h"
+#include "timer.h"
 
 struct libinput_source {
 	libinput_source_dispatch_t dispatch;
@@ -55,19 +56,19 @@ struct libinput_event_keyboard {
 	uint32_t time;
 	uint32_t key;
 	uint32_t seat_key_count;
-	enum libinput_keyboard_key_state state;
+	enum libinput_key_state state;
 };
 
 struct libinput_event_pointer {
 	struct libinput_event base;
 	uint32_t time;
-	li_fixed_t x;
-	li_fixed_t y;
+	double x;
+	double y;
 	uint32_t button;
 	uint32_t seat_button_count;
-	enum libinput_pointer_button_state state;
+	enum libinput_button_state state;
 	enum libinput_pointer_axis axis;
-	li_fixed_t value;
+	double value;
 };
 
 struct libinput_event_touch {
@@ -75,13 +76,13 @@ struct libinput_event_touch {
 	uint32_t time;
 	int32_t slot;
 	int32_t seat_slot;
-	li_fixed_t x;
-	li_fixed_t y;
+	double x;
+	double y;
 };
 
 static void
-libinput_default_log_func(enum libinput_log_priority priority,
-			  void *data,
+libinput_default_log_func(struct libinput *libinput,
+			  enum libinput_log_priority priority,
 			  const char *format, va_list args)
 {
 	const char *prefix;
@@ -97,48 +98,47 @@ libinput_default_log_func(enum libinput_log_priority priority,
 	vfprintf(stderr, format, args);
 }
 
-struct log_data {
-	enum libinput_log_priority priority;
-	libinput_log_handler handler;
-	void *user_data;
-};
-
-static struct log_data log_data = {
-	.priority = LIBINPUT_LOG_PRIORITY_ERROR,
-	.handler = libinput_default_log_func,
-	.user_data = NULL,
-};
+void
+log_msg_va(struct libinput *libinput,
+	   enum libinput_log_priority priority,
+	   const char *format,
+	   va_list args)
+{
+	if (libinput->log_handler &&
+	    libinput->log_priority <= priority)
+		libinput->log_handler(libinput, priority, format, args);
+}
 
 void
-log_msg(enum libinput_log_priority priority, const char *format, ...)
+log_msg(struct libinput *libinput,
+	enum libinput_log_priority priority,
+	const char *format, ...)
 {
 	va_list args;
 
-	if (log_data.handler && log_data.priority <= priority) {
-		va_start(args, format);
-		log_data.handler(priority, log_data.user_data, format, args);
-		va_end(args);
-	}
+	va_start(args, format);
+	log_msg_va(libinput, priority, format, args);
+	va_end(args);
 }
 
 LIBINPUT_EXPORT void
-libinput_log_set_priority(enum libinput_log_priority priority)
+libinput_log_set_priority(struct libinput *libinput,
+			  enum libinput_log_priority priority)
 {
-	log_data.priority = priority;
+	libinput->log_priority = priority;
 }
 
 LIBINPUT_EXPORT enum libinput_log_priority
-libinput_log_get_priority(void)
+libinput_log_get_priority(const struct libinput *libinput)
 {
-	return log_data.priority;
+	return libinput->log_priority;
 }
 
 LIBINPUT_EXPORT void
-libinput_log_set_handler(libinput_log_handler log_handler,
-			 void *user_data)
+libinput_log_set_handler(struct libinput *libinput,
+			 libinput_log_handler log_handler)
 {
-	log_data.handler = log_handler;
-	log_data.user_data = user_data;
+	libinput->log_handler = log_handler;
 }
 
 static void
@@ -277,7 +277,7 @@ libinput_event_keyboard_get_key(struct libinput_event_keyboard *event)
 	return event->key;
 }
 
-LIBINPUT_EXPORT enum libinput_keyboard_key_state
+LIBINPUT_EXPORT enum libinput_key_state
 libinput_event_keyboard_get_key_state(struct libinput_event_keyboard *event)
 {
 	return event->state;
@@ -296,31 +296,37 @@ libinput_event_pointer_get_time(struct libinput_event_pointer *event)
 	return event->time;
 }
 
-LIBINPUT_EXPORT li_fixed_t
+LIBINPUT_EXPORT double
 libinput_event_pointer_get_dx(struct libinput_event_pointer *event)
 {
 	return event->x;
 }
 
-LIBINPUT_EXPORT li_fixed_t
+LIBINPUT_EXPORT double
 libinput_event_pointer_get_dy(struct libinput_event_pointer *event)
 {
 	return event->y;
 }
 
-LIBINPUT_EXPORT li_fixed_t
+LIBINPUT_EXPORT double
 libinput_event_pointer_get_absolute_x(struct libinput_event_pointer *event)
 {
-	return event->x;
+	struct evdev_device *device =
+		(struct evdev_device *) event->base.device;
+
+	return evdev_convert_to_mm(device->abs.absinfo_x, event->x);
 }
 
-LIBINPUT_EXPORT li_fixed_t
+LIBINPUT_EXPORT double
 libinput_event_pointer_get_absolute_y(struct libinput_event_pointer *event)
 {
-	return event->y;
+	struct evdev_device *device =
+		(struct evdev_device *) event->base.device;
+
+	return evdev_convert_to_mm(device->abs.absinfo_y, event->y);
 }
 
-LIBINPUT_EXPORT li_fixed_t
+LIBINPUT_EXPORT double
 libinput_event_pointer_get_absolute_x_transformed(
 	struct libinput_event_pointer *event,
 	uint32_t width)
@@ -331,7 +337,7 @@ libinput_event_pointer_get_absolute_x_transformed(
 	return evdev_device_transform_x(device, event->x, width);
 }
 
-LIBINPUT_EXPORT li_fixed_t
+LIBINPUT_EXPORT double
 libinput_event_pointer_get_absolute_y_transformed(
 	struct libinput_event_pointer *event,
 	uint32_t height)
@@ -348,7 +354,7 @@ libinput_event_pointer_get_button(struct libinput_event_pointer *event)
 	return event->button;
 }
 
-LIBINPUT_EXPORT enum libinput_pointer_button_state
+LIBINPUT_EXPORT enum libinput_button_state
 libinput_event_pointer_get_button_state(struct libinput_event_pointer *event)
 {
 	return event->state;
@@ -367,7 +373,7 @@ libinput_event_pointer_get_axis(struct libinput_event_pointer *event)
 	return event->axis;
 }
 
-LIBINPUT_EXPORT li_fixed_t
+LIBINPUT_EXPORT double
 libinput_event_pointer_get_axis_value(struct libinput_event_pointer *event)
 {
 	return event->value;
@@ -391,13 +397,16 @@ libinput_event_touch_get_seat_slot(struct libinput_event_touch *event)
 	return event->seat_slot;
 }
 
-LIBINPUT_EXPORT li_fixed_t
+LIBINPUT_EXPORT double
 libinput_event_touch_get_x(struct libinput_event_touch *event)
 {
-	return event->x;
+	struct evdev_device *device =
+		(struct evdev_device *) event->base.device;
+
+	return evdev_convert_to_mm(device->abs.absinfo_x, event->x);
 }
 
-LIBINPUT_EXPORT li_fixed_t
+LIBINPUT_EXPORT double
 libinput_event_touch_get_x_transformed(struct libinput_event_touch *event,
 				       uint32_t width)
 {
@@ -407,7 +416,7 @@ libinput_event_touch_get_x_transformed(struct libinput_event_touch *event,
 	return evdev_device_transform_x(device, event->x, width);
 }
 
-LIBINPUT_EXPORT li_fixed_t
+LIBINPUT_EXPORT double
 libinput_event_touch_get_y_transformed(struct libinput_event_touch *event,
 				       uint32_t height)
 {
@@ -417,10 +426,13 @@ libinput_event_touch_get_y_transformed(struct libinput_event_touch *event,
 	return evdev_device_transform_y(device, event->y, height);
 }
 
-LIBINPUT_EXPORT li_fixed_t
+LIBINPUT_EXPORT double
 libinput_event_touch_get_y(struct libinput_event_touch *event)
 {
-	return event->y;
+	struct evdev_device *device =
+		(struct evdev_device *) event->base.device;
+
+	return evdev_convert_to_mm(device->abs.absinfo_y, event->y);
 }
 
 struct libinput_source *
@@ -445,7 +457,6 @@ libinput_add_fd(struct libinput *libinput,
 	ep.data.ptr = source;
 
 	if (epoll_ctl(libinput->epoll_fd, EPOLL_CTL_ADD, fd, &ep) < 0) {
-		close(source->fd);
 		free(source);
 		return NULL;
 	}
@@ -479,12 +490,21 @@ libinput_init(struct libinput *libinput,
 		return -1;
 	}
 
+	libinput->log_handler = libinput_default_log_func;
+	libinput->log_priority = LIBINPUT_LOG_PRIORITY_ERROR;
 	libinput->interface = interface;
 	libinput->interface_backend = interface_backend;
 	libinput->user_data = user_data;
+	libinput->refcount = 1;
 	list_init(&libinput->source_destroy_list);
 	list_init(&libinput->seat_list);
 
+	if (libinput_timer_subsys_init(libinput) != 0) {
+		free(libinput->events);
+		close(libinput->epoll_fd);
+		return -1;
+	}
+
 	return 0;
 }
 
@@ -504,15 +524,27 @@ libinput_drop_destroyed_sources(struct libinput *libinput)
 	list_init(&libinput->source_destroy_list);
 }
 
-LIBINPUT_EXPORT void
-libinput_destroy(struct libinput *libinput)
+LIBINPUT_EXPORT struct libinput *
+libinput_ref(struct libinput *libinput)
+{
+	libinput->refcount++;
+	return libinput;
+}
+
+LIBINPUT_EXPORT struct libinput *
+libinput_unref(struct libinput *libinput)
 {
 	struct libinput_event *event;
 	struct libinput_device *device, *next_device;
 	struct libinput_seat *seat, *next_seat;
 
 	if (libinput == NULL)
-		return;
+		return NULL;
+
+	assert(libinput->refcount > 0);
+	libinput->refcount--;
+	if (libinput->refcount > 0)
+		return libinput;
 
 	libinput_suspend(libinput);
 
@@ -521,8 +553,6 @@ libinput_destroy(struct libinput *libinput)
 	while ((event = libinput_get_event(libinput)))
 	       libinput_event_destroy(event);
 
-	libinput_drop_destroyed_sources(libinput);
-
 	free(libinput->events);
 
 	list_for_each_safe(seat, next_seat, &libinput->seat_list, link) {
@@ -534,8 +564,12 @@ libinput_destroy(struct libinput *libinput)
 		libinput_seat_destroy(seat);
 	}
 
+	libinput_timer_subsys_destroy(libinput);
+	libinput_drop_destroyed_sources(libinput);
 	close(libinput->epoll_fd);
 	free(libinput);
+
+	return NULL;
 }
 
 LIBINPUT_EXPORT void
@@ -581,10 +615,11 @@ libinput_seat_init(struct libinput_seat *seat,
 	list_insert(&libinput->seat_list, &seat->link);
 }
 
-LIBINPUT_EXPORT void
+LIBINPUT_EXPORT struct libinput_seat *
 libinput_seat_ref(struct libinput_seat *seat)
 {
 	seat->refcount++;
+	return seat;
 }
 
 static void
@@ -596,13 +631,17 @@ libinput_seat_destroy(struct libinput_seat *seat)
 	seat->destroy(seat);
 }
 
-LIBINPUT_EXPORT void
+LIBINPUT_EXPORT struct libinput_seat *
 libinput_seat_unref(struct libinput_seat *seat)
 {
 	assert(seat->refcount > 0);
 	seat->refcount--;
-	if (seat->refcount == 0)
+	if (seat->refcount == 0) {
 		libinput_seat_destroy(seat);
+		return NULL;
+	} else {
+		return seat;
+	}
 }
 
 LIBINPUT_EXPORT void
@@ -637,10 +676,11 @@ libinput_device_init(struct libinput_device *device,
 	device->refcount = 1;
 }
 
-LIBINPUT_EXPORT void
+LIBINPUT_EXPORT struct libinput_device *
 libinput_device_ref(struct libinput_device *device)
 {
 	device->refcount++;
+	return device;
 }
 
 static void
@@ -649,13 +689,17 @@ libinput_device_destroy(struct libinput_device *device)
 	evdev_device_destroy((struct evdev_device *) device);
 }
 
-LIBINPUT_EXPORT void
+LIBINPUT_EXPORT struct libinput_device *
 libinput_device_unref(struct libinput_device *device)
 {
 	assert(device->refcount > 0);
 	device->refcount--;
-	if (device->refcount == 0)
+	if (device->refcount == 0) {
 		libinput_device_destroy(device);
+		return NULL;
+	} else {
+		return device;
+	}
 }
 
 LIBINPUT_EXPORT int
@@ -691,14 +735,14 @@ libinput_dispatch(struct libinput *libinput)
 static uint32_t
 update_seat_key_count(struct libinput_seat *seat,
 		      int32_t key,
-		      enum libinput_keyboard_key_state state)
+		      enum libinput_key_state state)
 {
 	assert(key >= 0 && key <= KEY_MAX);
 
 	switch (state) {
-	case LIBINPUT_KEYBOARD_KEY_STATE_PRESSED:
+	case LIBINPUT_KEY_STATE_PRESSED:
 		return ++seat->button_count[key];
-	case LIBINPUT_KEYBOARD_KEY_STATE_RELEASED:
+	case LIBINPUT_KEY_STATE_RELEASED:
 		/* We might not have received the first PRESSED event. */
 		if (seat->button_count[key] == 0)
 			return 0;
@@ -712,14 +756,14 @@ update_seat_key_count(struct libinput_seat *seat,
 static uint32_t
 update_seat_button_count(struct libinput_seat *seat,
 			 int32_t button,
-			 enum libinput_pointer_button_state state)
+			 enum libinput_button_state state)
 {
 	assert(button >= 0 && button <= KEY_MAX);
 
 	switch (state) {
-	case LIBINPUT_POINTER_BUTTON_STATE_PRESSED:
+	case LIBINPUT_BUTTON_STATE_PRESSED:
 		return ++seat->button_count[button];
-	case LIBINPUT_POINTER_BUTTON_STATE_RELEASED:
+	case LIBINPUT_BUTTON_STATE_RELEASED:
 		/* We might not have received the first PRESSED event. */
 		if (seat->button_count[button] == 0)
 			return 0;
@@ -790,7 +834,7 @@ void
 keyboard_notify_key(struct libinput_device *device,
 		    uint32_t time,
 		    uint32_t key,
-		    enum libinput_keyboard_key_state state)
+		    enum libinput_key_state state)
 {
 	struct libinput_event_keyboard *key_event;
 	uint32_t seat_key_count;
@@ -816,8 +860,8 @@ keyboard_notify_key(struct libinput_device *device,
 void
 pointer_notify_motion(struct libinput_device *device,
 		      uint32_t time,
-		      li_fixed_t dx,
-		      li_fixed_t dy)
+		      double dx,
+		      double dy)
 {
 	struct libinput_event_pointer *motion_event;
 
@@ -839,8 +883,8 @@ pointer_notify_motion(struct libinput_device *device,
 void
 pointer_notify_motion_absolute(struct libinput_device *device,
 			       uint32_t time,
-			       li_fixed_t x,
-			       li_fixed_t y)
+			       double x,
+			       double y)
 {
 	struct libinput_event_pointer *motion_absolute_event;
 
@@ -863,7 +907,7 @@ void
 pointer_notify_button(struct libinput_device *device,
 		      uint32_t time,
 		      int32_t button,
-		      enum libinput_pointer_button_state state)
+		      enum libinput_button_state state)
 {
 	struct libinput_event_pointer *button_event;
 	int32_t seat_button_count;
@@ -892,7 +936,7 @@ void
 pointer_notify_axis(struct libinput_device *device,
 		    uint32_t time,
 		    enum libinput_pointer_axis axis,
-		    li_fixed_t value)
+		    double value)
 {
 	struct libinput_event_pointer *axis_event;
 
@@ -916,8 +960,8 @@ touch_notify_touch_down(struct libinput_device *device,
 			uint32_t time,
 			int32_t slot,
 			int32_t seat_slot,
-			li_fixed_t x,
-			li_fixed_t y)
+			double x,
+			double y)
 {
 	struct libinput_event_touch *touch_event;
 
@@ -943,8 +987,8 @@ touch_notify_touch_motion(struct libinput_device *device,
 			  uint32_t time,
 			  int32_t slot,
 			  int32_t seat_slot,
-			  li_fixed_t x,
-			  li_fixed_t y)
+			  double x,
+			  double y)
 {
 	struct libinput_event_touch *touch_event;
 
@@ -1007,7 +1051,6 @@ touch_notify_frame(struct libinput_device *device,
 			  &touch_event->base);
 }
 
-
 static void
 libinput_post_event(struct libinput *libinput,
 		    struct libinput_event *event)
@@ -1116,6 +1159,24 @@ libinput_device_get_sysname(struct libinput_device *device)
 	return evdev_device_get_sysname((struct evdev_device *) device);
 }
 
+LIBINPUT_EXPORT const char *
+libinput_device_get_name(struct libinput_device *device)
+{
+	return evdev_device_get_name((struct evdev_device *) device);
+}
+
+LIBINPUT_EXPORT unsigned int
+libinput_device_get_id_product(struct libinput_device *device)
+{
+	return evdev_device_get_id_product((struct evdev_device *) device);
+}
+
+LIBINPUT_EXPORT unsigned int
+libinput_device_get_id_vendor(struct libinput_device *device)
+{
+	return evdev_device_get_id_vendor((struct evdev_device *) device);
+}
+
 LIBINPUT_EXPORT const char *
 libinput_device_get_output_name(struct libinput_device *device)
 {
@@ -1159,6 +1220,16 @@ libinput_device_has_capability(struct libinput_device *device,
 					   capability);
 }
 
+LIBINPUT_EXPORT int
+libinput_device_get_size(struct libinput_device *device,
+			 double *width,
+			 double *height)
+{
+	return evdev_device_get_size((struct evdev_device *)device,
+				     width,
+				     height);
+}
+
 LIBINPUT_EXPORT struct libinput_event *
 libinput_event_device_notify_get_base_event(struct libinput_event_device_notify *event)
 {
@@ -1182,3 +1253,99 @@ libinput_event_touch_get_base_event(struct libinput_event_touch *event)
 {
 	return &event->base;
 }
+
+LIBINPUT_EXPORT const char *
+libinput_config_status_to_str(enum libinput_config_status status)
+{
+	const char *str = NULL;
+
+	switch(status) {
+	case LIBINPUT_CONFIG_STATUS_SUCCESS:
+		str = "Success";
+		break;
+	case LIBINPUT_CONFIG_STATUS_UNSUPPORTED:
+		str = "Unsupported configuration option";
+		break;
+	case LIBINPUT_CONFIG_STATUS_INVALID:
+		str = "Invalid argument range";
+		break;
+	}
+
+	return str;
+}
+
+LIBINPUT_EXPORT int
+libinput_device_config_tap_get_finger_count(struct libinput_device *device)
+{
+	return device->config.tap ? device->config.tap->count(device) : 0;
+}
+
+LIBINPUT_EXPORT enum libinput_config_status
+libinput_device_config_tap_set_enabled(struct libinput_device *device,
+				       enum libinput_config_tap_state enable)
+{
+	if (enable != LIBINPUT_CONFIG_TAP_ENABLED &&
+	    enable != LIBINPUT_CONFIG_TAP_DISABLED)
+		return LIBINPUT_CONFIG_STATUS_INVALID;
+
+	if (enable &&
+	    libinput_device_config_tap_get_finger_count(device) == 0)
+		return LIBINPUT_CONFIG_STATUS_UNSUPPORTED;
+
+	return device->config.tap->set_enabled(device, enable);
+}
+
+LIBINPUT_EXPORT enum libinput_config_tap_state
+libinput_device_config_tap_get_enabled(struct libinput_device *device)
+{
+	if (libinput_device_config_tap_get_finger_count(device) == 0)
+		return LIBINPUT_CONFIG_TAP_DISABLED;
+
+	return device->config.tap->get_enabled(device);
+}
+
+LIBINPUT_EXPORT enum libinput_config_tap_state
+libinput_device_config_tap_get_default_enabled(struct libinput_device *device)
+{
+	if (libinput_device_config_tap_get_finger_count(device) == 0)
+		return LIBINPUT_CONFIG_TAP_DISABLED;
+
+	return device->config.tap->get_default(device);
+}
+
+LIBINPUT_EXPORT int
+libinput_device_config_calibration_has_matrix(struct libinput_device *device)
+{
+	return device->config.calibration ?
+		device->config.calibration->has_matrix(device) : 0;
+}
+
+LIBINPUT_EXPORT enum libinput_config_status
+libinput_device_config_calibration_set_matrix(struct libinput_device *device,
+					      const float matrix[6])
+{
+	if (!libinput_device_config_calibration_has_matrix(device))
+		return LIBINPUT_CONFIG_STATUS_UNSUPPORTED;
+
+	return device->config.calibration->set_matrix(device, matrix);
+}
+
+LIBINPUT_EXPORT int
+libinput_device_config_calibration_get_matrix(struct libinput_device *device,
+					      float matrix[6])
+{
+	if (!libinput_device_config_calibration_has_matrix(device))
+		return 0;
+
+	return device->config.calibration->get_matrix(device, matrix);
+}
+
+LIBINPUT_EXPORT int
+libinput_device_config_calibration_get_default_matrix(struct libinput_device *device,
+						      float matrix[6])
+{
+	if (!libinput_device_config_calibration_has_matrix(device))
+		return 0;
+
+	return device->config.calibration->get_default_matrix(device, matrix);
+}
