@@ -49,6 +49,7 @@ device_added(struct udev_device *udev_device, struct udev_input *input)
 	const char *sysname;
 	const char *device_seat, *seat_name, *output_name;
 	const char *calibration_values;
+	float calibration[6];
 	struct udev_seat *seat;
 
 	device_seat = udev_device_get_property_value(udev_device, "ID_SEAT");
@@ -80,33 +81,34 @@ device_added(struct udev_device *udev_device, struct udev_input *input)
 	libinput_seat_unref(&seat->base);
 
 	if (device == EVDEV_UNHANDLED_DEVICE) {
-		log_info("not using input device '%s'.\n", devnode);
+		log_info(&input->base, "not using input device '%s'.\n", devnode);
 		return 0;
 	} else if (device == NULL) {
-		log_info("failed to create input device '%s'.\n", devnode);
+		log_info(&input->base, "failed to create input device '%s'.\n", devnode);
 		return 0;
 	}
 
 	calibration_values =
 		udev_device_get_property_value(udev_device,
-					       "WL_CALIBRATION");
+					       "LIBINPUT_CALIBRATION_MATRIX");
 
 	if (calibration_values && sscanf(calibration_values,
 					 "%f %f %f %f %f %f",
-					 &device->abs.calibration[0],
-					 &device->abs.calibration[1],
-					 &device->abs.calibration[2],
-					 &device->abs.calibration[3],
-					 &device->abs.calibration[4],
-					 &device->abs.calibration[5]) == 6) {
-		device->abs.apply_calibration = 1;
-		log_info("Applying calibration: %f %f %f %f %f %f\n",
-			 device->abs.calibration[0],
-			 device->abs.calibration[1],
-			 device->abs.calibration[2],
-			 device->abs.calibration[3],
-			 device->abs.calibration[4],
-			 device->abs.calibration[5]);
+					 &calibration[0],
+					 &calibration[1],
+					 &calibration[2],
+					 &calibration[3],
+					 &calibration[4],
+					 &calibration[5]) == 6) {
+		evdev_device_set_default_calibration(device, calibration);
+		log_info(&input->base,
+			 "Applying calibration: %f %f %f %f %f %f\n",
+			 calibration[0],
+			 calibration[1],
+			 calibration[2],
+			 calibration[3],
+			 calibration[4],
+			 calibration[5]);
 	}
 
 	output_name = udev_device_get_property_value(udev_device, "WL_OUTPUT");
@@ -128,7 +130,8 @@ device_removed(struct udev_device *udev_device, struct udev_input *input)
 		list_for_each_safe(device, next,
 				   &seat->base.devices_list, base.link) {
 			if (!strcmp(device->devnode, devnode)) {
-				log_info("input device %s, %s removed\n",
+				log_info(&input->base,
+					 "input device %s, %s removed\n",
 					 device->devname, device->devnode);
 				evdev_device_remove(device);
 				break;
@@ -214,7 +217,6 @@ udev_input_remove_devices(struct udev_input *input)
 	}
 }
 
-
 static void
 udev_input_disable(struct libinput *libinput)
 {
@@ -243,7 +245,8 @@ udev_input_enable(struct libinput *libinput)
 
 	input->udev_monitor = udev_monitor_new_from_netlink(udev, "udev");
 	if (!input->udev_monitor) {
-		log_info("udev: failed to create the udev monitor\n");
+		log_info(libinput,
+			 "udev: failed to create the udev monitor\n");
 		return -1;
 	}
 
@@ -251,7 +254,7 @@ udev_input_enable(struct libinput *libinput)
 			"input", NULL);
 
 	if (udev_monitor_enable_receiving(input->udev_monitor)) {
-		log_info("udev: failed to bind the udev monitor\n");
+		log_info(libinput, "udev: failed to bind the udev monitor\n");
 		udev_monitor_unref(input->udev_monitor);
 		input->udev_monitor = NULL;
 		return -1;
@@ -333,14 +336,13 @@ static const struct libinput_interface_backend interface_backend = {
 };
 
 LIBINPUT_EXPORT struct libinput *
-libinput_udev_create_for_seat(const struct libinput_interface *interface,
-			      void *user_data,
-			      struct udev *udev,
-			      const char *seat_id)
+libinput_udev_create_context(const struct libinput_interface *interface,
+			     void *user_data,
+			     struct udev *udev)
 {
 	struct udev_input *input;
 
-	if (!interface || !udev || !seat_id)
+	if (!interface || !udev)
 		return NULL;
 
 	input = zalloc(sizeof *input);
@@ -349,19 +351,36 @@ libinput_udev_create_for_seat(const struct libinput_interface *interface,
 
 	if (libinput_init(&input->base, interface,
 			  &interface_backend, user_data) != 0) {
+		libinput_unref(&input->base);
 		free(input);
 		return NULL;
 	}
 
 	input->udev = udev_ref(udev);
-	input->seat_id = strdup(seat_id);
-
-	if (udev_input_enable(&input->base) < 0) {
-		udev_unref(udev);
-		libinput_destroy(&input->base);
-		free(input);
-		return NULL;
-	}
 
 	return &input->base;
 }
+
+LIBINPUT_EXPORT int
+libinput_udev_assign_seat(struct libinput *libinput,
+			  const char *seat_id)
+{
+	struct udev_input *input = (struct udev_input*)libinput;
+
+	if (!seat_id)
+		return -1;
+	if (input->seat_id != NULL)
+		return -1;
+
+	if (libinput->interface_backend != &interface_backend) {
+		log_bug_client(libinput, "Mismatching backends.\n");
+		return -1;
+	}
+
+	input->seat_id = strdup(seat_id);
+
+	if (udev_input_enable(&input->base) < 0)
+		return -1;
+
+	return 0;
+}
