@@ -72,8 +72,8 @@ struct litest_test_device {
 };
 
 struct litest_device_interface {
-	void (*touch_down)(struct litest_device *d, unsigned int slot, int x, int y);
-	void (*touch_move)(struct litest_device *d, unsigned int slot, int x, int y);
+	void (*touch_down)(struct litest_device *d, unsigned int slot, double x, double y);
+	void (*touch_move)(struct litest_device *d, unsigned int slot, double x, double y);
 	void (*touch_up)(struct litest_device *d, unsigned int slot);
 
 	/**
@@ -93,7 +93,7 @@ struct litest_device_interface {
 };
 
 void litest_set_current_device(struct litest_device *device);
-int litest_scale(const struct litest_device *d, unsigned int axis, int val);
+int litest_scale(const struct litest_device *d, unsigned int axis, double val);
 void litest_generic_device_teardown(void);
 
 #endif
