@@ -24,6 +24,7 @@
 #include "config.h"
 #endif
 
+#include <assert.h>
 #include <check.h>
 #include <errno.h>
 #include <fcntl.h>
@@ -34,7 +35,7 @@
 #include <stdlib.h>
 #include <time.h>
 #include <unistd.h>
-#include <linux/input.h>
+#include "linux/input.h"
 #include <sys/ptrace.h>
 #include <sys/timerfd.h>
 #include <sys/wait.h>
@@ -44,6 +45,7 @@
 #include "libinput-util.h"
 
 static int in_debugger = -1;
+static int verbose = 0;
 
 struct test {
 	struct list node;
@@ -78,23 +80,28 @@ void litest_generic_device_teardown(void)
 extern struct litest_test_device litest_keyboard_device;
 extern struct litest_test_device litest_synaptics_clickpad_device;
 extern struct litest_test_device litest_synaptics_touchpad_device;
+extern struct litest_test_device litest_synaptics_t440_device;
 extern struct litest_test_device litest_trackpoint_device;
 extern struct litest_test_device litest_bcm5974_device;
 extern struct litest_test_device litest_mouse_device;
 extern struct litest_test_device litest_wacom_touch_device;
+extern struct litest_test_device litest_alps_device;
+extern struct litest_test_device litest_generic_singletouch_device;
 
 struct litest_test_device* devices[] = {
 	&litest_synaptics_clickpad_device,
 	&litest_synaptics_touchpad_device,
+	&litest_synaptics_t440_device,
 	&litest_keyboard_device,
 	&litest_trackpoint_device,
 	&litest_bcm5974_device,
 	&litest_mouse_device,
 	&litest_wacom_touch_device,
+	&litest_alps_device,
+	&litest_generic_singletouch_device,
 	NULL,
 };
 
-
 static struct list all_tests;
 
 static void
@@ -176,11 +183,8 @@ litest_add_no_device(const char *name, void *func)
 	litest_add(name, func, LITEST_DISABLE_DEVICE, LITEST_DISABLE_DEVICE);
 }
 
-void
-litest_add(const char *name,
-	   void *func,
-	   enum litest_device_feature required,
-	   enum litest_device_feature excluded)
+static struct suite *
+get_suite(const char *name)
 {
 	struct suite *s;
 
@@ -188,10 +192,8 @@ litest_add(const char *name,
 		list_init(&all_tests);
 
 	list_for_each(s, &all_tests, node) {
-		if (strcmp(s->name, name) == 0) {
-			litest_add_tcase(s, func, required, excluded);
-			return;
-		}
+		if (strcmp(s->name, name) == 0)
+			return s;
 	}
 
 	s = zalloc(sizeof(*s));
@@ -200,10 +202,41 @@ litest_add(const char *name,
 
 	list_init(&s->tests);
 	list_insert(&all_tests, &s->node);
-	litest_add_tcase(s, func, required, excluded);
+
+	return s;
 }
 
-int is_debugger_attached()
+void
+litest_add(const char *name,
+	   void *func,
+	   enum litest_device_feature required,
+	   enum litest_device_feature excluded)
+{
+	litest_add_tcase(get_suite(name), func, required, excluded);
+}
+
+void
+litest_add_for_device(const char *name,
+		      void *func,
+		      enum litest_device_type type)
+{
+	struct suite *s;
+	struct litest_test_device **dev = devices;
+
+	s = get_suite(name);
+	while (*dev) {
+		if ((*dev)->type == type) {
+			litest_add_tcase_for_device(s, func, *dev);
+			return;
+		}
+		dev++;
+	}
+
+	ck_abort_msg("Invalid test device type");
+}
+
+static int
+is_debugger_attached(void)
 {
 	int status;
 	int rc;
@@ -231,7 +264,6 @@ int is_debugger_attached()
 	return rc;
 }
 
-
 static void
 litest_list_tests(struct list *tests)
 {
@@ -246,8 +278,44 @@ litest_list_tests(struct list *tests)
 	}
 }
 
+static void
+litest_log_handler(struct libinput *libinput,
+		   enum libinput_log_priority pri,
+		   const char *format,
+		   va_list args)
+{
+	const char *priority = NULL;
+
+	switch(pri) {
+	case LIBINPUT_LOG_PRIORITY_INFO: priority = "info"; break;
+	case LIBINPUT_LOG_PRIORITY_ERROR: priority = "error"; break;
+	case LIBINPUT_LOG_PRIORITY_DEBUG: priority = "debug"; break;
+	}
+
+	fprintf(stderr, "litest %s: ", priority);
+	vfprintf(stderr, format, args);
+}
+
+static int
+open_restricted(const char *path, int flags, void *userdata)
+{
+	return open(path, flags);
+}
+
+static void
+close_restricted(int fd, void *userdata)
+{
+	close(fd);
+}
+
+struct libinput_interface interface = {
+	.open_restricted = open_restricted,
+	.close_restricted = close_restricted,
+};
+
 static const struct option opts[] = {
 	{ "list", 0, 0, 'l' },
+	{ "verbose", 0, 0, 'v' },
 	{ 0, 0, 0, 0}
 };
 
@@ -281,6 +349,9 @@ litest_run(int argc, char **argv) {
 			case 'l':
 				litest_list_tests(&all_tests);
 				return 0;
+			case 'v':
+				verbose = 1;
+				break;
 			default:
 				fprintf(stderr, "usage: %s [--list]\n", argv[0]);
 				return 1;
@@ -288,7 +359,7 @@ litest_run(int argc, char **argv) {
 		}
 	}
 
-	srunner_run_all(sr, CK_NORMAL);
+	srunner_run_all(sr, CK_ENV);
 	failed = srunner_ntests_failed(sr);
 	srunner_free(sr);
 
@@ -309,30 +380,12 @@ litest_run(int argc, char **argv) {
 	return failed;
 }
 
-static int
-open_restricted(const char *path, int flags, void *userdata)
-{
-	return open(path, flags);
-}
-
-static void
-close_restricted(int fd, void *userdata)
-{
-	close(fd);
-}
-
-const struct libinput_interface interface = {
-	.open_restricted = open_restricted,
-	.close_restricted = close_restricted,
-};
-
-
 static struct input_absinfo *
 merge_absinfo(const struct input_absinfo *orig,
 	      const struct input_absinfo *override)
 {
 	struct input_absinfo *abs;
-	int nelem, i;
+	unsigned int nelem, i;
 	size_t sz = ABS_MAX + 1;
 
 	if (!orig)
@@ -366,7 +419,7 @@ static int*
 merge_events(const int *orig, const int *override)
 {
 	int *events;
-	int nelem, i;
+	unsigned int nelem, i;
 	size_t sz = KEY_MAX * 3;
 
 	if (!orig)
@@ -396,7 +449,6 @@ merge_events(const int *orig, const int *override)
 	return events;
 }
 
-
 static struct litest_device *
 litest_create(enum litest_device_type which,
 	      const char *name_override,
@@ -457,6 +509,11 @@ litest_create_context(void)
 	struct libinput *libinput =
 		libinput_path_create_context(&interface, NULL);
 	ck_assert_notnull(libinput);
+
+	libinput_log_set_handler(libinput, litest_log_handler);
+	if (verbose)
+		libinput_log_set_priority(libinput, LIBINPUT_LOG_PRIORITY_DEBUG);
+
 	return libinput;
 }
 
@@ -546,10 +603,12 @@ litest_delete_device(struct litest_device *d)
 		return;
 
 	libinput_device_unref(d->libinput_device);
+	libinput_path_remove_device(d->libinput_device);
 	if (d->owns_context)
-		libinput_destroy(d->libinput);
+		libinput_unref(d->libinput);
 	libevdev_free(d->evdev);
 	libevdev_uinput_destroy(d->uinput);
+	free(d->private);
 	memset(d,0, sizeof(*d));
 	free(d);
 }
@@ -558,13 +617,14 @@ void
 litest_event(struct litest_device *d, unsigned int type,
 	     unsigned int code, int value)
 {
-	libevdev_uinput_write_event(d->uinput, type, code, value);
+	int ret = libevdev_uinput_write_event(d->uinput, type, code, value);
+	ck_assert_int_eq(ret, 0);
 }
 
-static int
-auto_assign_value(struct litest_device *d,
-		  const struct input_event *ev,
-		  int slot, int x, int y)
+int
+litest_auto_assign_value(struct litest_device *d,
+			 const struct input_event *ev,
+			 int slot, double x, double y)
 {
 	static int tracking_id;
 	int value = ev->value;
@@ -592,12 +652,27 @@ auto_assign_value(struct litest_device *d,
 	return value;
 }
 
+static void
+send_btntool(struct litest_device *d)
+{
+	litest_event(d, EV_KEY, BTN_TOUCH, d->ntouches_down != 0);
+	litest_event(d, EV_KEY, BTN_TOOL_FINGER, d->ntouches_down == 1);
+	litest_event(d, EV_KEY, BTN_TOOL_DOUBLETAP, d->ntouches_down == 2);
+	litest_event(d, EV_KEY, BTN_TOOL_TRIPLETAP, d->ntouches_down == 3);
+	litest_event(d, EV_KEY, BTN_TOOL_QUADTAP, d->ntouches_down == 4);
+	litest_event(d, EV_KEY, BTN_TOOL_QUINTTAP, d->ntouches_down == 5);
+}
 
 void
-litest_touch_down(struct litest_device *d, unsigned int slot, int x, int y)
+litest_touch_down(struct litest_device *d, unsigned int slot,
+		  double x, double y)
 {
 	struct input_event *ev;
 
+	assert(++d->ntouches_down > 0);
+
+	send_btntool(d);
+
 	if (d->interface->touch_down) {
 		d->interface->touch_down(d, slot, x, y);
 		return;
@@ -605,7 +680,7 @@ litest_touch_down(struct litest_device *d, unsigned int slot, int x, int y)
 
 	ev = d->interface->touch_down_events;
 	while (ev && (int16_t)ev->type != -1 && (int16_t)ev->code != -1) {
-		int value = auto_assign_value(d, ev, slot, x, y);
+		int value = litest_auto_assign_value(d, ev, slot, x, y);
 		litest_event(d, ev->type, ev->code, value);
 		ev++;
 	}
@@ -622,6 +697,10 @@ litest_touch_up(struct litest_device *d, unsigned int slot)
 		{ .type = -1, .code = -1 }
 	};
 
+	assert(--d->ntouches_down >= 0);
+
+	send_btntool(d);
+
 	if (d->interface->touch_up) {
 		d->interface->touch_up(d, slot);
 		return;
@@ -631,14 +710,15 @@ litest_touch_up(struct litest_device *d, unsigned int slot)
 		ev = up;
 
 	while (ev && (int16_t)ev->type != -1 && (int16_t)ev->code != -1) {
-		int value = auto_assign_value(d, ev, slot, 0, 0);
+		int value = litest_auto_assign_value(d, ev, slot, 0, 0);
 		litest_event(d, ev->type, ev->code, value);
 		ev++;
 	}
 }
 
 void
-litest_touch_move(struct litest_device *d, unsigned int slot, int x, int y)
+litest_touch_move(struct litest_device *d, unsigned int slot,
+		  double x, double y)
 {
 	struct input_event *ev;
 
@@ -649,7 +729,7 @@ litest_touch_move(struct litest_device *d, unsigned int slot, int x, int y)
 
 	ev = d->interface->touch_move_events;
 	while (ev && (int16_t)ev->type != -1 && (int16_t)ev->code != -1) {
-		int value = auto_assign_value(d, ev, slot, x, y);
+		int value = litest_auto_assign_value(d, ev, slot, x, y);
 		litest_event(d, ev->type, ev->code, value);
 		ev++;
 	}
@@ -658,8 +738,8 @@ litest_touch_move(struct litest_device *d, unsigned int slot, int x, int y)
 void
 litest_touch_move_to(struct litest_device *d,
 		     unsigned int slot,
-		     int x_from, int y_from,
-		     int x_to, int y_to,
+		     double x_from, double y_from,
+		     double x_to, double y_to,
 		     int steps)
 {
 	for (int i = 0; i < steps - 1; i++)
@@ -689,7 +769,8 @@ litest_keyboard_key(struct litest_device *d, unsigned int key, bool is_press)
 	litest_button_click(d, key, is_press);
 }
 
-int litest_scale(const struct litest_device *d, unsigned int axis, int val)
+int
+litest_scale(const struct litest_device *d, unsigned int axis, double val)
 {
 	int min, max;
 	ck_assert_int_ge(val, 0);
@@ -701,6 +782,53 @@ int litest_scale(const struct litest_device *d, unsigned int axis, int val)
 	return (max - min) * val/100.0 + min;
 }
 
+void
+litest_wait_for_event(struct libinput *li)
+{
+	return litest_wait_for_event_of_type(li, -1);
+}
+
+void
+litest_wait_for_event_of_type(struct libinput *li, ...)
+{
+	va_list args;
+	enum libinput_event_type types[32] = {LIBINPUT_EVENT_NONE};
+	size_t ntypes = 0;
+	enum libinput_event_type type;
+
+	va_start(args, li);
+	type = va_arg(args, int);
+	while ((int)type != -1) {
+		assert(type > 0);
+		assert(ntypes < ARRAY_LENGTH(types));
+		types[ntypes++] = type;
+		type = va_arg(args, int);
+	}
+	va_end(args);
+
+	while (1) {
+		size_t i;
+		struct libinput_event *event;
+
+		while ((type = libinput_next_event_type(li)) == LIBINPUT_EVENT_NONE) {
+			msleep(10);
+			libinput_dispatch(li);
+		}
+
+		/* no event mask means wait for any event */
+		if (ntypes == 0)
+			return;
+
+		for (i = 0; i < ntypes; i++) {
+			if (type == types[i])
+				return;
+		}
+
+		event = libinput_get_event(li);
+		libinput_event_destroy(event);
+	}
+}
+
 void
 litest_drain_events(struct libinput *li)
 {
@@ -713,16 +841,78 @@ litest_drain_events(struct libinput *li)
 	}
 }
 
+static void
+litest_print_event(struct libinput_event *event)
+{
+	struct libinput_event_pointer *p;
+	struct libinput_device *dev;
+	enum libinput_event_type type;
+	double x, y;
+
+	dev = libinput_event_get_device(event);
+	type = libinput_event_get_type(event);
+
+	fprintf(stderr,
+		"device %s type %d ",
+		libinput_device_get_sysname(dev),
+		type);
+	switch (type) {
+	case LIBINPUT_EVENT_POINTER_MOTION:
+		p = libinput_event_get_pointer_event(event);
+		x = libinput_event_pointer_get_dx(p);
+		y = libinput_event_pointer_get_dy(p);
+		fprintf(stderr, "motion: %.2f/%.2f", x, y);
+		break;
+	case LIBINPUT_EVENT_POINTER_MOTION_ABSOLUTE:
+		p = libinput_event_get_pointer_event(event);
+		x = libinput_event_pointer_get_absolute_x(p);
+		y = libinput_event_pointer_get_absolute_y(p);
+		fprintf(stderr, "motion: %.2f/%.2f", x, y);
+		break;
+	case LIBINPUT_EVENT_POINTER_BUTTON:
+		p = libinput_event_get_pointer_event(event);
+		fprintf(stderr,
+			"button: %d state %d",
+			libinput_event_pointer_get_button(p),
+			libinput_event_pointer_get_button_state(p));
+		break;
+	default:
+		break;
+	}
+
+	fprintf(stderr, "\n");
+}
+
+void
+litest_assert_empty_queue(struct libinput *li)
+{
+	bool empty_queue = true;
+	struct libinput_event *event;
+
+	libinput_dispatch(li);
+	while ((event = libinput_get_event(li))) {
+		empty_queue = false;
+		fprintf(stderr,
+			"Unexpected event: ");
+		litest_print_event(event);
+		libinput_event_destroy(event);
+		libinput_dispatch(li);
+	}
+
+	ck_assert(empty_queue);
+}
+
 struct libevdev_uinput *
 litest_create_uinput_device_from_description(const char *name,
 					     const struct input_id *id,
-					     const struct input_absinfo *abs,
+					     const struct input_absinfo *abs_info,
 					     const int *events)
 {
 	struct libevdev_uinput *uinput;
 	struct libevdev *dev;
 	int type, code;
-	int rc;
+	int rc, fd;
+	const struct input_absinfo *abs;
 	const struct input_absinfo default_abs = {
 		.value = 0,
 		.minimum = 0,
@@ -731,17 +921,21 @@ litest_create_uinput_device_from_description(const char *name,
 		.flat = 0,
 		.resolution = 100
 	};
+	char buf[512];
+	const char *devnode;
 
 	dev = libevdev_new();
 	ck_assert(dev != NULL);
 
-	libevdev_set_name(dev, name);
+	snprintf(buf, sizeof(buf), "litest %s", name);
+	libevdev_set_name(dev, buf);
 	if (id) {
 		libevdev_set_id_bustype(dev, id->bustype);
 		libevdev_set_id_vendor(dev, id->vendor);
 		libevdev_set_id_product(dev, id->product);
 	}
 
+	abs = abs_info;
 	while (abs && abs->value != -1) {
 		rc = libevdev_enable_event_code(dev, EV_ABS,
 						abs->value, abs);
@@ -755,8 +949,6 @@ litest_create_uinput_device_from_description(const char *name,
 		if (type == INPUT_PROP_MAX) {
 			rc = libevdev_enable_property(dev, code);
 		} else {
-			if (type != EV_SYN)
-				ck_assert(!libevdev_has_event_code(dev, type, code));
 			rc = libevdev_enable_event_code(dev, type, code,
 							type == EV_ABS ? &default_abs : NULL);
 		}
@@ -770,6 +962,31 @@ litest_create_uinput_device_from_description(const char *name,
 
 	libevdev_free(dev);
 
+	/* uinput does not yet support setting the resolution, so we set it
+	 * afterwards. This is of course racy as hell but the way we
+	 * _generally_ use this function by the time libinput uses the
+	 * device, we're finished here */
+
+	devnode = libevdev_uinput_get_devnode(uinput);
+	ck_assert_notnull(devnode);
+	fd = open(devnode, O_RDONLY);
+	ck_assert_int_gt(fd, -1);
+	rc = libevdev_new_from_fd(fd, &dev);
+	ck_assert_int_eq(rc, 0);
+
+	abs = abs_info;
+	while (abs && abs->value != -1) {
+		if (abs->resolution != 0) {
+			rc = libevdev_kernel_set_abs_info(dev,
+							  abs->value,
+							  abs);
+			ck_assert_int_eq(rc, 0);
+		}
+		abs++;
+	}
+	close(fd);
+	libevdev_free(dev);
+
 	return uinput;
 }
 
