@@ -26,6 +26,7 @@
 #include <errno.h>
 #include <fcntl.h>
 #include <libinput.h>
+#include <libinput-util.h>
 #include <unistd.h>
 
 #include "litest.h"
@@ -68,8 +69,8 @@ create_simple_test_device(const char *name, ...)
 
 	va_start(args, name);
 
-	while ((type = va_arg(args, unsigned int)) != -1 &&
-	       (code = va_arg(args, unsigned int)) != -1) {
+	while ((type = va_arg(args, unsigned int)) != (unsigned int)-1 &&
+	       (code = va_arg(args, unsigned int)) != (unsigned int)-1) {
 		const struct input_absinfo *a = NULL;
 		if (type == EV_ABS)
 			a = &abs;
@@ -94,7 +95,7 @@ START_TEST(event_conversion_device_notify)
 	struct libinput_event *event;
 	int device_added = 0, device_removed = 0;
 
-	uinput = create_simple_test_device("test device",
+	uinput = create_simple_test_device("litest test device",
 					   EV_REL, REL_X,
 					   EV_REL, REL_Y,
 					   EV_KEY, BTN_LEFT,
@@ -133,7 +134,7 @@ START_TEST(event_conversion_device_notify)
 		libinput_event_destroy(event);
 	}
 
-	libinput_destroy(li);
+	libinput_unref(li);
 	libevdev_uinput_destroy(uinput);
 
 	ck_assert_int_gt(device_added, 0);
@@ -148,7 +149,7 @@ START_TEST(event_conversion_pointer)
 	struct libinput_event *event;
 	int motion = 0, button = 0;
 
-	uinput = create_simple_test_device("test device",
+	uinput = create_simple_test_device("litest test device",
 					   EV_REL, REL_X,
 					   EV_REL, REL_Y,
 					   EV_KEY, BTN_LEFT,
@@ -158,6 +159,11 @@ START_TEST(event_conversion_pointer)
 	li = libinput_path_create_context(&simple_interface, NULL);
 	libinput_path_add_device(li, libevdev_uinput_get_devnode(uinput));
 
+	/* Queue at least two relative motion events as the first one may
+	 * be absorbed by the pointer acceleration filter. */
+	libevdev_uinput_write_event(uinput, EV_REL, REL_X, -1);
+	libevdev_uinput_write_event(uinput, EV_REL, REL_Y, -1);
+	libevdev_uinput_write_event(uinput, EV_SYN, SYN_REPORT, 0);
 	libevdev_uinput_write_event(uinput, EV_REL, REL_X, -1);
 	libevdev_uinput_write_event(uinput, EV_REL, REL_Y, -1);
 	libevdev_uinput_write_event(uinput, EV_KEY, BTN_LEFT, 1);
@@ -189,7 +195,7 @@ START_TEST(event_conversion_pointer)
 		libinput_event_destroy(event);
 	}
 
-	libinput_destroy(li);
+	libinput_unref(li);
 	libevdev_uinput_destroy(uinput);
 
 	ck_assert_int_gt(motion, 0);
@@ -204,7 +210,7 @@ START_TEST(event_conversion_pointer_abs)
 	struct libinput_event *event;
 	int motion = 0, button = 0;
 
-	uinput = create_simple_test_device("test device",
+	uinput = create_simple_test_device("litest test device",
 					   EV_ABS, ABS_X,
 					   EV_ABS, ABS_Y,
 					   EV_KEY, BTN_LEFT,
@@ -249,7 +255,7 @@ START_TEST(event_conversion_pointer_abs)
 		libinput_event_destroy(event);
 	}
 
-	libinput_destroy(li);
+	libinput_unref(li);
 	libevdev_uinput_destroy(uinput);
 
 	ck_assert_int_gt(motion, 0);
@@ -264,7 +270,7 @@ START_TEST(event_conversion_key)
 	struct libinput_event *event;
 	int key = 0;
 
-	uinput = create_simple_test_device("test device",
+	uinput = create_simple_test_device("litest test device",
 					   EV_KEY, KEY_A,
 					   EV_KEY, KEY_B,
 					   -1, -1);
@@ -299,7 +305,7 @@ START_TEST(event_conversion_key)
 		libinput_event_destroy(event);
 	}
 
-	libinput_destroy(li);
+	libinput_unref(li);
 	libevdev_uinput_destroy(uinput);
 
 	ck_assert_int_gt(key, 0);
@@ -313,7 +319,7 @@ START_TEST(event_conversion_touch)
 	struct libinput_event *event;
 	int touch = 0;
 
-	uinput = create_simple_test_device("test device",
+	uinput = create_simple_test_device("litest test device",
 					   EV_KEY, BTN_TOUCH,
 					   EV_ABS, ABS_X,
 					   EV_ABS, ABS_Y,
@@ -359,19 +365,159 @@ START_TEST(event_conversion_touch)
 		libinput_event_destroy(event);
 	}
 
-	libinput_destroy(li);
+	libinput_unref(li);
 	libevdev_uinput_destroy(uinput);
 
 	ck_assert_int_gt(touch, 0);
 }
 END_TEST
 
+START_TEST(context_ref_counting)
+{
+	struct libinput *li;
+
+	/* These tests rely on valgrind to detect memory leak and use after
+	 * free errors. */
+
+	li = libinput_path_create_context(&simple_interface, NULL);
+	ck_assert_notnull(li);
+	ck_assert_ptr_eq(libinput_unref(li), NULL);
+
+	li = libinput_path_create_context(&simple_interface, NULL);
+	ck_assert_notnull(li);
+	ck_assert_ptr_eq(libinput_ref(li), li);
+	ck_assert_ptr_eq(libinput_unref(li), li);
+	ck_assert_ptr_eq(libinput_unref(li), NULL);
+}
+END_TEST
+
+START_TEST(device_ids)
+{
+	struct litest_device *dev = litest_current_device();
+	const char *name;
+	unsigned int pid, vid;
+
+	name = libevdev_get_name(dev->evdev);
+	pid = libevdev_get_id_product(dev->evdev);
+	vid = libevdev_get_id_vendor(dev->evdev);
+
+	ck_assert_str_eq(name,
+			 libinput_device_get_name(dev->libinput_device));
+	ck_assert_int_eq(pid,
+			 libinput_device_get_id_product(dev->libinput_device));
+	ck_assert_int_eq(vid,
+			 libinput_device_get_id_vendor(dev->libinput_device));
+}
+END_TEST
+
+START_TEST(config_status_string)
+{
+	const char *strs[3];
+	const char *invalid;
+	size_t i, j;
+
+	strs[0] = libinput_config_status_to_str(LIBINPUT_CONFIG_STATUS_SUCCESS);
+	strs[1] = libinput_config_status_to_str(LIBINPUT_CONFIG_STATUS_UNSUPPORTED);
+	strs[2] = libinput_config_status_to_str(LIBINPUT_CONFIG_STATUS_INVALID);
+
+	for (i = 0; i < ARRAY_LENGTH(strs) - 1; i++)
+		for (j = i + 1; j < ARRAY_LENGTH(strs); j++)
+			ck_assert_str_ne(strs[i], strs[j]);
+
+	invalid = libinput_config_status_to_str(LIBINPUT_CONFIG_STATUS_INVALID + 1);
+	ck_assert(invalid == NULL);
+	invalid = libinput_config_status_to_str(LIBINPUT_CONFIG_STATUS_SUCCESS - 1);
+	ck_assert(invalid == NULL);
+}
+END_TEST
+
+START_TEST(matrix_helpers)
+{
+	struct matrix m1, m2, m3;
+	float f[6] = { 1, 2, 3, 4, 5, 6 };
+	int x, y;
+	int row, col;
+
+	matrix_init_identity(&m1);
+
+	for (row = 0; row < 3; row++) {
+		for (col = 0; col < 3; col++) {
+			ck_assert_int_eq(m1.val[row][col],
+					 (row == col) ? 1 : 0);
+		}
+	}
+	ck_assert(matrix_is_identity(&m1));
+
+	matrix_from_farray6(&m2, f);
+	ck_assert_int_eq(m2.val[0][0], 1);
+	ck_assert_int_eq(m2.val[0][1], 2);
+	ck_assert_int_eq(m2.val[0][2], 3);
+	ck_assert_int_eq(m2.val[1][0], 4);
+	ck_assert_int_eq(m2.val[1][1], 5);
+	ck_assert_int_eq(m2.val[1][2], 6);
+	ck_assert_int_eq(m2.val[2][0], 0);
+	ck_assert_int_eq(m2.val[2][1], 0);
+	ck_assert_int_eq(m2.val[2][2], 1);
+
+	x = 100;
+	y = 5;
+	matrix_mult_vec(&m1, &x, &y);
+	ck_assert_int_eq(x, 100);
+	ck_assert_int_eq(y, 5);
+
+	matrix_mult(&m3, &m1, &m1);
+	ck_assert(matrix_is_identity(&m3));
+
+	matrix_init_scale(&m2, 2, 4);
+	ck_assert_int_eq(m2.val[0][0], 2);
+	ck_assert_int_eq(m2.val[0][1], 0);
+	ck_assert_int_eq(m2.val[0][2], 0);
+	ck_assert_int_eq(m2.val[1][0], 0);
+	ck_assert_int_eq(m2.val[1][1], 4);
+	ck_assert_int_eq(m2.val[1][2], 0);
+	ck_assert_int_eq(m2.val[2][0], 0);
+	ck_assert_int_eq(m2.val[2][1], 0);
+	ck_assert_int_eq(m2.val[2][2], 1);
+
+	matrix_mult_vec(&m2, &x, &y);
+	ck_assert_int_eq(x, 200);
+	ck_assert_int_eq(y, 20);
+
+	matrix_init_translate(&m2, 10, 100);
+	ck_assert_int_eq(m2.val[0][0], 1);
+	ck_assert_int_eq(m2.val[0][1], 0);
+	ck_assert_int_eq(m2.val[0][2], 10);
+	ck_assert_int_eq(m2.val[1][0], 0);
+	ck_assert_int_eq(m2.val[1][1], 1);
+	ck_assert_int_eq(m2.val[1][2], 100);
+	ck_assert_int_eq(m2.val[2][0], 0);
+	ck_assert_int_eq(m2.val[2][1], 0);
+	ck_assert_int_eq(m2.val[2][2], 1);
+
+	matrix_mult_vec(&m2, &x, &y);
+	ck_assert_int_eq(x, 210);
+	ck_assert_int_eq(y, 120);
+
+	matrix_to_farray6(&m2, f);
+	ck_assert_int_eq(f[0], 1);
+	ck_assert_int_eq(f[1], 0);
+	ck_assert_int_eq(f[2], 10);
+	ck_assert_int_eq(f[3], 0);
+	ck_assert_int_eq(f[4], 1);
+	ck_assert_int_eq(f[5], 100);
+}
+END_TEST
+
 int main (int argc, char **argv) {
 	litest_add_no_device("events:conversion", event_conversion_device_notify);
 	litest_add_no_device("events:conversion", event_conversion_pointer);
 	litest_add_no_device("events:conversion", event_conversion_pointer_abs);
 	litest_add_no_device("events:conversion", event_conversion_key);
 	litest_add_no_device("events:conversion", event_conversion_touch);
+	litest_add_no_device("context:refcount", context_ref_counting);
+	litest_add("device:id", device_ids, LITEST_ANY, LITEST_ANY);
+	litest_add_no_device("config:status string", config_status_string);
 
+	litest_add_no_device("misc:matrix", matrix_helpers);
 	return litest_run(argc, argv);
 }
