@@ -53,20 +53,18 @@ const struct libinput_interface simple_interface = {
 	.close_restricted = close_restricted,
 };
 
-
 START_TEST(path_create_NULL)
 {
 	struct libinput *li;
-	const struct libinput_interface interface;
 
 	open_func_count = 0;
 	close_func_count = 0;
 
 	li = libinput_path_create_context(NULL, NULL);
 	ck_assert(li == NULL);
-	li = libinput_path_create_context(&interface, NULL);
+	li = libinput_path_create_context(&simple_interface, NULL);
 	ck_assert(li != NULL);
-	libinput_destroy(li);
+	libinput_unref(li);
 
 	ck_assert_int_eq(open_func_count, 0);
 	ck_assert_int_eq(close_func_count, 0);
@@ -93,7 +91,7 @@ START_TEST(path_create_invalid)
 	ck_assert_int_eq(open_func_count, 0);
 	ck_assert_int_eq(close_func_count, 0);
 
-	libinput_destroy(li);
+	libinput_unref(li);
 	ck_assert_int_eq(close_func_count, 0);
 
 	open_func_count = 0;
@@ -127,7 +125,7 @@ START_TEST(path_create_destroy)
 	ck_assert_int_eq(open_func_count, 1);
 
 	libevdev_uinput_destroy(uinput);
-	libinput_destroy(li);
+	libinput_unref(li);
 	ck_assert_int_eq(close_func_count, 1);
 
 	open_func_count = 0;
@@ -158,7 +156,7 @@ START_TEST(path_added_seat)
 	ck_assert(seat != NULL);
 
 	seat_name = libinput_seat_get_logical_name(seat);
-	ck_assert_int_eq(strcmp(seat_name, "default"), 0);
+	ck_assert_str_eq(seat_name, "default");
 
 	libinput_event_destroy(event);
 }
@@ -237,7 +235,7 @@ START_TEST(path_add_device)
 		libinput_event_destroy(event);
 	}
 
-	ck_assert_int_eq(strcmp(sysname1, sysname2), 0);
+	ck_assert_str_eq(sysname1, sysname2);
 
 	libinput_event_destroy(event);
 }
@@ -245,12 +243,11 @@ END_TEST
 
 START_TEST(path_add_invalid_path)
 {
-	struct litest_device *dev = litest_current_device();
-	struct libinput *li = dev->libinput;
+	struct libinput *li;
 	struct libinput_event *event;
 	struct libinput_device *device;
 
-	litest_drain_events(li);
+	li = litest_create_context();
 
 	device = libinput_path_add_device(li, "/tmp/");
 	ck_assert(device == NULL);
@@ -259,6 +256,8 @@ START_TEST(path_add_invalid_path)
 
 	while ((event = libinput_get_event(li)))
 		ck_abort();
+
+	libinput_unref(li);
 }
 END_TEST
 
@@ -373,7 +372,7 @@ START_TEST(path_suspend)
 	libinput_resume(li);
 
 	libevdev_uinput_destroy(uinput);
-	libinput_destroy(li);
+	libinput_unref(li);
 
 	open_func_count = 0;
 	close_func_count = 0;
@@ -407,7 +406,7 @@ START_TEST(path_double_suspend)
 	libinput_resume(li);
 
 	libevdev_uinput_destroy(uinput);
-	libinput_destroy(li);
+	libinput_unref(li);
 
 	open_func_count = 0;
 	close_func_count = 0;
@@ -441,7 +440,7 @@ START_TEST(path_double_resume)
 	libinput_resume(li);
 
 	libevdev_uinput_destroy(uinput);
-	libinput_destroy(li);
+	libinput_unref(li);
 
 	open_func_count = 0;
 	close_func_count = 0;
@@ -493,7 +492,6 @@ START_TEST(path_add_device_suspend_resume)
 
 	ck_assert_int_eq(nevents, 2);
 
-
 	libinput_suspend(li);
 	libinput_dispatch(li);
 
@@ -524,7 +522,7 @@ START_TEST(path_add_device_suspend_resume)
 
 	libevdev_uinput_destroy(uinput1);
 	libevdev_uinput_destroy(uinput2);
-	libinput_destroy(li);
+	libinput_unref(li);
 
 	open_func_count = 0;
 	close_func_count = 0;
@@ -577,7 +575,6 @@ START_TEST(path_add_device_suspend_resume_fail)
 
 	ck_assert_int_eq(nevents, 2);
 
-
 	libinput_suspend(li);
 	libinput_dispatch(li);
 
@@ -615,7 +612,7 @@ START_TEST(path_add_device_suspend_resume_fail)
 	ck_assert_int_eq(nevents, 2);
 
 	libevdev_uinput_destroy(uinput2);
-	libinput_destroy(li);
+	libinput_unref(li);
 
 	open_func_count = 0;
 	close_func_count = 0;
@@ -668,7 +665,6 @@ START_TEST(path_add_device_suspend_resume_remove_device)
 
 	ck_assert_int_eq(nevents, 2);
 
-
 	libinput_suspend(li);
 	libinput_dispatch(li);
 
@@ -705,7 +701,7 @@ START_TEST(path_add_device_suspend_resume_remove_device)
 	ck_assert_int_eq(nevents, 1);
 
 	libevdev_uinput_destroy(uinput1);
-	libinput_destroy(li);
+	libinput_unref(li);
 
 	open_func_count = 0;
 	close_func_count = 0;
@@ -791,32 +787,32 @@ START_TEST(path_seat_recycle)
 
 	ck_assert(found == 1);
 
-	libinput_destroy(li);
+	libinput_unref(li);
 
 	libevdev_uinput_destroy(uinput);
 }
 END_TEST
 
-int main (int argc, char **argv) {
-
-	litest_add("path:create", path_create_NULL, LITEST_ANY, LITEST_ANY);
-	litest_add("path:create", path_create_invalid, LITEST_ANY, LITEST_ANY);
-	litest_add("path:create", path_create_destroy, LITEST_ANY, LITEST_ANY);
-	litest_add("path:suspend", path_suspend, LITEST_ANY, LITEST_ANY);
-	litest_add("path:suspend", path_double_suspend, LITEST_ANY, LITEST_ANY);
-	litest_add("path:suspend", path_double_resume, LITEST_ANY, LITEST_ANY);
-	litest_add("path:suspend", path_add_device_suspend_resume, LITEST_ANY, LITEST_ANY);
-	litest_add("path:suspend", path_add_device_suspend_resume_fail, LITEST_ANY, LITEST_ANY);
-	litest_add("path:suspend", path_add_device_suspend_resume_remove_device, LITEST_ANY, LITEST_ANY);
-	litest_add("path:seat events", path_added_seat, LITEST_ANY, LITEST_ANY);
+int
+main(int argc, char **argv)
+{
+	litest_add_no_device("path:create", path_create_NULL);
+	litest_add_no_device("path:create", path_create_invalid);
+	litest_add_no_device("path:create", path_create_destroy);
+	litest_add_no_device("path:suspend", path_suspend);
+	litest_add_no_device("path:suspend", path_double_suspend);
+	litest_add_no_device("path:suspend", path_double_resume);
+	litest_add_no_device("path:suspend", path_add_device_suspend_resume);
+	litest_add_no_device("path:suspend", path_add_device_suspend_resume_fail);
+	litest_add_no_device("path:suspend", path_add_device_suspend_resume_remove_device);
+	litest_add_for_device("path:seat events", path_added_seat, LITEST_SYNAPTICS_CLICKPAD);
 	litest_add("path:device events", path_added_device, LITEST_ANY, LITEST_ANY);
 	litest_add("path:device events", path_device_sysname, LITEST_ANY, LITEST_ANY);
-	litest_add("path:device events", path_add_device, LITEST_ANY, LITEST_ANY);
-	litest_add("path:device events", path_add_invalid_path, LITEST_ANY, LITEST_ANY);
-	litest_add("path:device events", path_remove_device, LITEST_ANY, LITEST_ANY);
-	litest_add("path:device events", path_double_remove_device, LITEST_ANY, LITEST_ANY);
-	litest_add("path:seat", path_seat_recycle,
-		   LITEST_DISABLE_DEVICE, LITEST_DISABLE_DEVICE);
+	litest_add_for_device("path:device events", path_add_device, LITEST_SYNAPTICS_CLICKPAD);
+	litest_add_no_device("path:device events", path_add_invalid_path);
+	litest_add_for_device("path:device events", path_remove_device, LITEST_SYNAPTICS_CLICKPAD);
+	litest_add_for_device("path:device events", path_double_remove_device, LITEST_SYNAPTICS_CLICKPAD);
+	litest_add_no_device("path:seat", path_seat_recycle);
 
 	return litest_run(argc, argv);
 }
