@@ -30,7 +30,8 @@
 #include <string.h>
 #include <time.h>
 #include <unistd.h>
-#include <linux/input.h>
+#include <libudev.h>
+#include "linux/input.h"
 #include <sys/ioctl.h>
 #include <sys/signalfd.h>
 
@@ -125,11 +126,20 @@ close_restricted(int fd, void *user_data)
 	close(fd);
 }
 
-const static struct libinput_interface interface = {
+static const struct libinput_interface interface = {
 	.open_restricted = open_restricted,
 	.close_restricted = close_restricted,
 };
 
+static void
+log_handler(struct libinput *li,
+	    enum libinput_log_priority priority,
+	    const char *format,
+	    va_list args)
+{
+	vprintf(format, args);
+}
+
 static int
 open_udev(struct libinput **li)
 {
@@ -139,12 +149,23 @@ open_udev(struct libinput **li)
 		return 1;
 	}
 
-	*li = libinput_udev_create_for_seat(&interface, NULL, udev, seat);
+	*li = libinput_udev_create_context(&interface, NULL, udev);
 	if (!*li) {
 		fprintf(stderr, "Failed to initialize context from udev\n");
 		return 1;
 	}
 
+	if (verbose) {
+		libinput_log_set_handler(*li, log_handler);
+		libinput_log_set_priority(*li, LIBINPUT_LOG_PRIORITY_DEBUG);
+	}
+
+	if (libinput_udev_assign_seat(*li, seat)) {
+		fprintf(stderr, "Failed to set seat\n");
+		libinput_unref(*li);
+		return 1;
+	}
+
 	return 0;
 }
 
@@ -159,10 +180,15 @@ open_device(struct libinput **li, const char *path)
 		return 1;
 	}
 
+	if (verbose) {
+		libinput_log_set_handler(*li, log_handler);
+		libinput_log_set_priority(*li, LIBINPUT_LOG_PRIORITY_DEBUG);
+	}
+
 	device = libinput_path_add_device(*li, path);
 	if (!device) {
 		fprintf(stderr, "Failed to initialized device %s\n", path);
-		libinput_destroy(*li);
+		libinput_unref(*li);
 		return 1;
 	}
 
@@ -173,7 +199,7 @@ static void
 print_event_header(struct libinput_event *ev)
 {
 	struct libinput_device *dev = libinput_event_get_device(ev);
-	const char *type;
+	const char *type = NULL;
 
 	switch(libinput_event_get_type(ev)) {
 	case LIBINPUT_EVENT_NONE:
@@ -230,66 +256,68 @@ print_device_notify(struct libinput_event *ev)
 {
 	struct libinput_device *dev = libinput_event_get_device(ev);
 	struct libinput_seat *seat = libinput_device_get_seat(dev);
+	double w, h;
 
-	printf("%s	%s\n",
+	printf("%s	%s",
 	       libinput_seat_get_physical_name(seat),
 	       libinput_seat_get_logical_name(seat));
+
+	if (libinput_device_get_size(dev, &w, &h) == 0)
+		printf("\tsize %.2f/%.2fmm", w, h);
+
+	printf("\n");
 }
 
 static void
 print_key_event(struct libinput_event *ev)
 {
 	struct libinput_event_keyboard *k = libinput_event_get_keyboard_event(ev);
-	enum libinput_keyboard_key_state state;
+	enum libinput_key_state state;
 
 	print_event_time(libinput_event_keyboard_get_time(k));
 	state = libinput_event_keyboard_get_key_state(k);
 	printf("%d %s\n",
 	       libinput_event_keyboard_get_key(k),
-	       state == LIBINPUT_KEYBOARD_KEY_STATE_PRESSED ? "pressed" : "released");
+	       state == LIBINPUT_KEY_STATE_PRESSED ? "pressed" : "released");
 }
 
 static void
 print_motion_event(struct libinput_event *ev)
 {
 	struct libinput_event_pointer *p = libinput_event_get_pointer_event(ev);
-	li_fixed_t x =	libinput_event_pointer_get_dx(p),
-		   y = libinput_event_pointer_get_dy(p);
+	double x = libinput_event_pointer_get_dx(p);
+	double y = libinput_event_pointer_get_dy(p);
 
 	print_event_time(libinput_event_pointer_get_time(p));
 
-	printf("%6.2f/%6.2f\n",
-	       li_fixed_to_double(x),
-	       li_fixed_to_double(y));
+	printf("%6.2f/%6.2f\n", x, y);
 }
 
 static void
 print_absmotion_event(struct libinput_event *ev)
 {
 	struct libinput_event_pointer *p = libinput_event_get_pointer_event(ev);
-	li_fixed_t x = libinput_event_pointer_get_absolute_x_transformed(
+	double x = libinput_event_pointer_get_absolute_x_transformed(
 		p, screen_width);
-	li_fixed_t y = libinput_event_pointer_get_absolute_y_transformed(
+	double y = libinput_event_pointer_get_absolute_y_transformed(
 		p, screen_height);
 
 	print_event_time(libinput_event_pointer_get_time(p));
-	printf("%6.2f/%6.2f\n",
-	       li_fixed_to_double(x),
-	       li_fixed_to_double(y));
+	printf("%6.2f/%6.2f\n", x, y);
 }
 
 static void
 print_button_event(struct libinput_event *ev)
 {
 	struct libinput_event_pointer *p = libinput_event_get_pointer_event(ev);
-	enum libinput_pointer_button_state state;
+	enum libinput_button_state state;
 
 	print_event_time(libinput_event_pointer_get_time(p));
 
 	state = libinput_event_pointer_get_button_state(p);
 	printf("%3d %s, seat count: %u\n",
 	       libinput_event_pointer_get_button(p),
-	       state == LIBINPUT_POINTER_BUTTON_STATE_PRESSED ? "pressed" : "released",
+	       state == LIBINPUT_BUTTON_STATE_PRESSED ? "pressed" : "released",
 	       libinput_event_pointer_get_seat_button_count(p));
 }
 
@@ -299,13 +327,13 @@ print_axis_event(struct libinput_event *ev)
 	struct libinput_event_pointer *p = libinput_event_get_pointer_event(ev);
 	enum libinput_pointer_axis axis = libinput_event_pointer_get_axis(p);
 	const char *ax;
-	li_fixed_t val;
+	double val;
 
 	switch (axis) {
-	case LIBINPUT_POINTER_AXIS_VERTICAL_SCROLL:
+	case LIBINPUT_POINTER_AXIS_SCROLL_VERTICAL:
 		ax = "vscroll";
 		break;
-	case LIBINPUT_POINTER_AXIS_HORIZONTAL_SCROLL:
+	case LIBINPUT_POINTER_AXIS_SCROLL_HORIZONTAL:
 		ax = "hscroll";
 		break;
 	default:
@@ -314,8 +342,7 @@ print_axis_event(struct libinput_event *ev)
 
 	print_event_time(libinput_event_pointer_get_time(p));
 	val = libinput_event_pointer_get_axis_value(p);
-	printf("%s %.2f\n",
-	       ax, li_fixed_to_double(val));
+	printf("%s %.2f\n", ax, val);
 }
 
 static void
@@ -331,16 +358,18 @@ static void
 print_touch_event_with_coords(struct libinput_event *ev)
 {
 	struct libinput_event_touch *t = libinput_event_get_touch_event(ev);
-	li_fixed_t x = libinput_event_touch_get_x_transformed(t, screen_width),
-		   y = libinput_event_touch_get_y_transformed(t, screen_height);
+	double x = libinput_event_touch_get_x_transformed(t, screen_width);
+	double y = libinput_event_touch_get_y_transformed(t, screen_height);
+	double xmm = libinput_event_touch_get_x(t);
+	double ymm = libinput_event_touch_get_y(t);
 
 	print_event_time(libinput_event_touch_get_time(t));
 
-	printf("%d (%d) %5.2f/%5.2f\n",
+	printf("%d (%d) %5.2f/%5.2f (%5.2f/%5.2fmm)\n",
 	       libinput_event_touch_get_slot(t),
 	       libinput_event_touch_get_seat_slot(t),
-	       li_fixed_to_double(x),
-	       li_fixed_to_double(y));
+	       x, y,
+	       xmm, ymm);
 }
 
 static int
@@ -399,7 +428,7 @@ handle_and_print_events(struct libinput *li)
 	return rc;
 }
 
-void
+static void
 mainloop(struct libinput *li)
 {
 	struct pollfd fds[2];
@@ -437,15 +466,6 @@ mainloop(struct libinput *li)
 	close(fds[1].fd);
 }
 
-static void
-log_handler(enum libinput_log_priority priority,
-	    void *user_data,
-	    const char *format,
-	    va_list args)
-{
-	vprintf(format, args);
-}
-
 int
 main(int argc, char **argv)
 {
@@ -455,11 +475,6 @@ main(int argc, char **argv)
 	if (parse_args(argc, argv))
 		return 1;
 
-	if (verbose) {
-		libinput_log_set_handler(log_handler, NULL);
-		libinput_log_set_priority(LIBINPUT_LOG_PRIORITY_DEBUG);
-	}
-
 	if (mode == MODE_UDEV) {
 		if (open_udev(&li))
 			return 1;
@@ -474,6 +489,7 @@ main(int argc, char **argv)
 
 	mainloop(li);
 
+	libinput_unref(li);
 	if (udev)
 		udev_unref(udev);
 
